#include "nds_asm.h"
#include "equates.h"
#include "armsdtmac.h"
#include "ARMZ80/z80mac.h"

	.arm
	.align 4
	.section .text

	.global CPU_reset
	.global run
	.global default_scanlinehook
	.global waitMaskIn
	.global waitMaskOut



;@----------------------------------------------------------------------------
run:		;@ return after 1 frame
;@----------------------------------------------------------------------------

	ldrb r0,waitCountIn
	add r0,r0,#1
	strb r0,waitCountIn
	ldrb r1,waitMaskIn
	ands r0,r0,r1
	bxne lr
	stmfd sp!,{r4-r11,lr}

	ldr z80optbl,=Z80_optable
	b run_start
;@----------------------------------------------------------------------------
;@cycles ran out
;@----------------------------------------------------------------------------
line0:
	add r2,z80optbl,#z80_regs
	stmia r2,{z80f-z80pc,z80sp}	;@ save Z80 state

	ldrb r0,waitCountOut
	add r0,r0,#1
	strb r0,waitCountOut
	ldrb r1,waitMaskOut
	ands r0,r0,r1
	ldmeqfd sp!,{r4-r11,lr}		;@ exit here if doing single frame:
	bxeq lr						;@ return to rommenu()

run_start:
	adr lr,line0x				;@ return here after doing L/R + SEL/START
	ldr r0,=EMUinput
	ldr r0,[r0]

	ldr r2,=ystart
	ldrb r1,[r2]
	tst r0,#0x200				;@ L?
	subnes r1,#1
	movmi r1,#0
	tst r0,#0x100				;@ R?
	addne r1,#1
	cmp r1,#224-192
	movpl r1,#224-192
	strb r1,[r2]

line0x:
	bl refreshEMUjoypads		;@ Z=1 if communication ok
;@	bne waitformulti			;@ waiting on other GBA..

	bl newframe					;@ display update

	adr_ r0,z80_regs
	ldmia r0,{z80f-z80pc,z80sp}	;@ restore Z80 state

	ldr r0,cyclesperscanline
	add cycles,cycles,r0
	adr r0,line1_to_120
	str_ r0,z80_nexttimeout
	str_ r0,z80_nexttimeout_
	mov r1,#0					;@ Scanline

	mov r11,r11					;@ No$GBA breakpoint.
	ldr pc,scanlinehook
line1_to_120: ;@------------------------
	ldr r0,cyclesperscanline
	add cycles,cycles,r0

	ldr r0,=scanline
	ldr r1,[r0]
	add r1,r1,#1
	str r1,[r0]
	cmp r1,#120
	ldrmi pc,scanlinehook
;@--------------------------------------------- between 120 and 121

	adr addy,line121_to_240
	str_ addy,z80_nexttimeout
	str_ addy,z80_nexttimeout_
	ldr pc,scanlinehook
line121_to_240: ;@------------------------
	ldr r0,cyclesperscanline
	add cycles,cycles,r0

	ldr r0,=scanline
	ldr r1,[r0]
	add r1,r1,#1
	str r1,[r0]
	cmp r1,#240
	ldrmi pc,scanlinehook

	adr addy,line241
	str_ addy,z80_nexttimeout
	str_ addy,z80_nexttimeout_
	ldr pc,scanlinehook
line241: ;@------------------------
	bl endframe					;@ display update


/* [ DEBUG
	mov r1,#REG_BASE			;@ darken screen during EMU vblank
	mov r0,#0x00f1
	strh r0,[r1,#REG_BLDCNT]
	ldrh r0,[r1,#REG_VCOUNT]
	mov r1,#19
	bl debug_
 ]*/

	ldr r0,=irqControl
	ldrb r0,[r0]
	tst r0,#6					;@ IRQ enabled?, gb=4, goemon=2
	beq novbirq

	ldrb_ r0,z80_if_
	orr r0,r0,#0x01
	strb_ r0,z80_if_

novbirq:
	mov r0,#24*CYCLE
	add cycles,cycles,r0

	ldr r0,=scanline
	mov r1,#241
	str r1,[r0]

	adr addy,VBL_Hook
	str_ addy,z80_nexttimeout
	str_ addy,z80_nexttimeout_
	b _GO
VBL_Hook:
	ldr r0,cyclesperscanline
	sub r0,r0,#24*CYCLE
	add cycles,cycles,r0

	adr addy,line242_to_end
	str_ addy,z80_nexttimeout
	str_ addy,z80_nexttimeout_
	ldr pc,scanlinehook
line242_to_end: ;@------------------------
	ldr r0,cyclesperscanline
	add cycles,cycles,r0

	ldr r0,=scanline
	ldr r1,[r0]
	add r1,r1,#1
	str r1,[r0]
	cmp r1,#264					;@ last scanline
	ldrmi pc,scanlinehook

	adr addy,line0
	str_ addy,z80_nexttimeout
	str_ addy,z80_nexttimeout_

	ldr r0,=frametotal
	ldr r1,[r0]
	add r1,r1,#1
	str r1,[r0]

	ldr r0,=scanline
	ldr r1,[r0]
	add r1,r1,#1
	str r1,[r0]

	ldr pc,scanlinehook

;@----------------------------------------------------------------------------
default_scanlinehook:
checkScanlineIRQ:
	ldr r0,=scanline
	ldr r1,[r0]
	ldrb r0,scanlinemod
	tst r1,r0					;@ gb=0xf, goemon=0x1f
	bne noScanlineIRQ
	ldr r2,=irqControl
	ldrb r2,[r2]
	tst r2,#1					;@ NMI enabled?
	bne Z80_nmi
noScanlineIRQ:
	b checkMasterIRQ

;@----------------------------------------------------------------------------
scanlinehook:		.long 0
cyclesperscanline:	.long 0
waitCountIn:		.byte 0
waitCountOut:		.byte 0
waitMaskIn:			.byte 1
waitMaskOut:		.byte 0
scanlinemod:		.byte 0
					.byte 0
					.byte 0,0

;@----------------------------------------------------------------------------
CPU_reset:		;@ called by loadcart/resetGame
;@----------------------------------------------------------------------------
	str lr,[sp,#-4]!

	ldr r0,=default_scanlinehook
	ldr r1,=scanlinehook
	str r0,[r1]

	ldr r0,=goemon
	ldr r0,[r0]
	cmp r0,#0
	moveq r0,#0x0f
	movne r0,#0x1f
	ldr r1,=scanlinemod
	strb r0,[r1]
;@---Speed - 3.072MHz / 30Hz		;Green Beret.
	ldreq r0,=391*CYCLE
;@---Speed - 3.072MHz / 60Hz		;Goemon.
	ldrne r0,=195*CYCLE
	ldr r1,=cyclesperscanline
	str r0,[r1]
	bl Z80_reset
	ldr lr,[sp],#4
	bx lr

;@----------------------------------------------------------------------------


