#include <nds.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "gui.h"
#include "filehandling.h"
#include "unzip/unzipnds.h"
#include "emubase.h"
#include "ARMZ80/version.h"
#include "SN76496/version.h"

#define ARRSIZE(xxxx) (sizeof((xxxx))/sizeof((xxxx)[0]))
#define EMUVERSION "V0.2"

//asm variables
extern u8 g_dipswitch0;			// cart.s
extern u8 g_dipswitch1;			// cart.s
extern u8 g_dipswitch2;			// cart.s
extern int goemon;				// cart.s
extern u8 g_scaling;			// gfx.s
extern u8 g_flicker;			// gfx.s
extern u8 g_twitch;				// gfx.s
extern u8 g_gfxmask;			// gfx.s
extern u8 g_gammavalue;			// gfx.s
extern u32 joycfg;				// io.s
extern u32 EMUinput;

//asm functions
extern void CPU_reset(void);	//cpu.s
extern void loadcart(int,int);	// cart.s
extern void paletteInit(void);
extern void paletteTxAll(void);

extern int pauseEmulation;		// main.c
extern int sleepTime;			// main.c
extern uint16* map0sub;			// main.c

extern void waitVBlank(void);	// main.c
extern void setEmuSpeed(int speed);	// main.c

const fptr fnlist0[]={uiDummy};
const fptr fnlist1[]={ui8,loadState,saveState,saveSettings,resetGame};
const fptr fnlist2[]={ui4,ui5,ui6,ui7};
const fptr fnlist3[]={uiDummy};
const fptr fnlist4[]={autoBSet,autoASet,controllerSet,swapABSet};
const fptr fnlist5[]={scalingSet,flickSet,brightSet,bgrlayerSet,sprlayerSet};
const fptr fnlist6[]={difficultSet,coinaSet,coinbSet,livesSet,bonusSet,cabinetSet,demoSet,flipSet,uprightSet,serviceSet};
const fptr fnlist7[]={speedSet,fpsSet,sleepSet,autostateSet,autopauseGameSet,powerSaveSet};
const fptr fnlist8[]={loadGame,loadGame,loadGame,loadGame};
const fptr *const fnlistX[]={fnlist0,fnlist1,fnlist2,fnlist3,fnlist4,fnlist5,fnlist6,fnlist7,fnlist8};
const u8 menuXitems[]={ARRSIZE(fnlist0),ARRSIZE(fnlist1),ARRSIZE(fnlist2),ARRSIZE(fnlist3),ARRSIZE(fnlist4),ARRSIZE(fnlist5),ARRSIZE(fnlist6),ARRSIZE(fnlist7),ARRSIZE(fnlist8)};
const fptr drawuiX[]={uiNull,uiFile,uiOptions,uiAbout,uiController,uiDisplay,uiDipswitches,uiSettings,uiLoadGame};
const u8 menuXback[]={0,0,0,0,2,2,2,2,1};

int menuItemRow=0;
int selected=0;
int selectedMenu=0;
int selectedMain=0;
int lastMainMenu=1;
int mainUIPosition=0;

u8 autoA=0,autoB=0;				//0=off, 1=on, 2=R
int emuSettings=1;


char *const autotxt[]={"OFF","ON","with R"};
char *const speedtxt[]={"Normal","200%","Max","50%"};
char *const sleeptxt[]={"5min","10min","30min","OFF"};
char *const brightxt[]={"I","II","III","IIII","IIIII"};
char *const ctrltxt[]={"1P","2P"};
char *const disptxt[]={"UNSCALED","SCALED"};
char *const flicktxt[]={"No Flicker","Flicker"};
char *const difftxt[]={"Easy","Normal","Hard","Very Hard"};
char *const cointxt[]={ "1Coin 1Credit","1Coin 2Credits","1Coin 3Credits","1Coin 4Credits",\
						"1Coin 5Credits","1Coin 6Credits","1Coin 7Credits","2Coins 1Credit",\
						"2Coins 3Credits","2Coins 5Credits","3Coins 1Credit","3Coins 2Credits",\
						"3Coins 4Credits","4Coins 1Credit","4Coins 3Credits","Free Play"};
char *const livestxt[]={"2","3","5","7"};
char *const bonustxt[]={"30K 70K 70K+","40K 80K 80K+","50K 100K 100K+","50K 200K 200K+"};
char *const cabtxt[]={"Cocktail","Upright"};
char *const singletxt[]={"Single","Dual"};


void setupGUI() {
	keysSetRepeat(25, 4);	// delay, repeat.
	cls(0);
	uiNull();
	selected = 0;		// menu item 0.
}

void uiNull() {
	cls(1);
	drawtext("      Touch screen or",10,0);
	drawtext("     press L+R for menu.",11,0);
	pauseEmulation = 0;
	if(emuSettings & 2) {
		powerOff(PM_BACKLIGHT_BOTTOM);
		powerOff(POWER_2D_B);
	}
}

void uiFile() {
	cls(1);
	drawTabs();
	menuItemRow = 0;
	drawMenuItem("Load Game");
	drawMenuItem("Load State");
	drawMenuItem("Save State");
	drawMenuItem("Save Settings");
	drawMenuItem("Reset Game");
}

void uiOptions() {
	cls(1);
	drawTabs();
	menuItemRow = 0;
	drawMenuItem("Controller");
	drawMenuItem("Display");
	drawMenuItem("DipSwitches");
	drawMenuItem("Settings");
}

void uiAbout() {
	cls(1);
	drawTabs();
	drawtext(" Select: Insert coin",4,0);
	drawtext(" Start:  Start button",5,0);
	drawtext(" DPad:   Move character",6,0);
	drawtext(" Up:     Jump/Climb",7,0);
	drawtext(" Down:   Crouch",8,0);
	drawtext(" B:      Knife attack",9,0);
	drawtext(" A:      Special attack",10,0);

	drawtext(" ARMZ80       " ARMZ80VERSION,21,0);
	drawtext(" ARMSN76496   " ARMSNVERSION,22,0);
	drawtext(" GreenBeretDS " EMUVERSION,23,0);
}

void uiController() {
	cls(1);
	drawTabs();
	drawtext(" Controller Settings",3,0);
	menuItemRow = 0;
	drawSubItem("B autofire: ",autotxt[autoB]);
	drawSubItem("A autofire: ",autotxt[autoA]);
	drawSubItem("Controller: ",ctrltxt[(joycfg>>29)&1]);
	drawSubItem("Swap A-B:   ",autotxt[(joycfg>>10)&1]);
}

void uiDisplay() {
	cls(1);
	drawTabs();
	drawtext(" Display Settings",3,0);
	menuItemRow = 0;
	drawSubItem("Display: ",disptxt[g_scaling]);
	drawSubItem("Scaling: ",flicktxt[g_flicker]);
	drawSubItem("Gamma: ",brightxt[g_gammavalue]);
	drawSubItem("Disable Background: ",autotxt[g_gfxmask&1]);
	drawSubItem("Disable Sprites: ",autotxt[(g_gfxmask>>4)&1]);
}

void uiDipswitches() {
	cls(1);
	drawTabs();
	drawtext(" Dipswitch Settings",3,0);
	menuItemRow = 0;
	drawSubItem("Difficulty: ",difftxt[(g_dipswitch1>>5)&3]);
	drawSubItem("Coin A: ",cointxt[g_dipswitch0 & 0xF]);
	drawSubItem("Coin B: ",cointxt[(g_dipswitch0>>4) & 0xF]);
	drawSubItem("Lives: ",livestxt[g_dipswitch1 & 3]);
	drawSubItem("Bonus: ",bonustxt[(g_dipswitch1>>3)&3]);
	drawSubItem("Cabinet: ",cabtxt[(g_dipswitch1>>2)&1]);
	drawSubItem("Demo Sound: ",autotxt[(g_dipswitch1>>7)&1]);
	drawSubItem("Flip Screen: ",autotxt[g_dipswitch2&1]);
	drawSubItem("Upright Controls: ",singletxt[(g_dipswitch2>>1)&1]);
	drawSubItem("Service Mode: ",autotxt[(g_dipswitch2>>2)&1]);
//	drawSubItem("CoinCounter1:","        ");
//	int2str(g_coin0,s);
//	drawSubItem("CoinCounter2:","        ");
//	int2str(g_coin1,s);
}

void uiSettings() {
	cls(1);
	drawTabs();
	drawtext(" Settings",3,0);
	menuItemRow = 0;
	drawSubItem("Speed: ",speedtxt[(emuSettings>>6)&3]);
	drawSubItem("FPS-Meter: ",autotxt[(emuSettings>>3)&1]);
	drawSubItem("Autosleep: ",sleeptxt[(emuSettings>>4)&3]);
	drawSubItem("Autoload state: ",autotxt[(emuSettings>>2)&1]);
	drawSubItem("Autopause game: ",autotxt[emuSettings&1]);
	drawSubItem("Powersave 2nd screen: ",autotxt[(emuSettings>>1)&1]);
}

void uiLoadGame() {
	cls(1);
	drawTabs();
	drawtext(" Load game",3,0);
	menuItemRow = 0;
	drawMenuItem(" Green Beret");
	drawMenuItem(" Rush'n Attack (US)");
	drawMenuItem(" Green Beret (bootleg)");
	drawMenuItem(" Mr. Goemon (Japan)");
}

char menuTopRow[]={   0x82,0x83, ' ',0x81,0x82,0x82,0x82,0x82,0x83, ' ',0x81,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x83, ' ',0x81,0x82,0x82,0x82,0x82,0x82,0x83,0};
char menuMiddleRow[]={ 'X',0x85, ' ',0x84, 'F', 'i', 'l', 'e',0x85, ' ',0x84, 'O', 'p', 't', 'i', 'o', 'n', 's',0x85, ' ',0x84, 'A', 'b', 'o', 'u', 't',0x85,0};
char menuBottomRow[]={ ' ',0x8A, ' ',0x89, ' ', ' ', ' ', ' ',0x8A, ' ',0x89, ' ', ' ', ' ', ' ', ' ', ' ', ' ',0x8A, ' ',0x89, ' ', ' ', ' ', ' ', ' ',0x8A,0};
char menuFrontRow[]={ 0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0};

char tabTopFile[]={0x81,0x82,0x82,0x82,0x82,0x83,0};
char tabMidFile[]={0x84, 'F', 'i', 'l', 'e',0x85,0};
char tabBotFile[]={0x89, ' ', ' ', ' ', ' ',0x8A,0};

char tabTopOptions[]={0x81,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x83,0};
char tabMidOptions[]={0x84, 'O', 'p', 't', 'i', 'o', 'n', 's',0x85,0};
char tabBotOptions[]={0x89, ' ', ' ', ' ', ' ', ' ', ' ', ' ',0x8A,0};

char tabTopAbout[]={0x81,0x82,0x82,0x82,0x82,0x82,0x83,0};
char tabMidAbout[]={0x84, 'A', 'b', 'o', 'u', 't',0x85,0};
char tabBotAbout[]={0x89, ' ', ' ', ' ', ' ', ' ',0x8A,0};

void drawTabs() {
	drawbtext(menuTopRow,0,1);
	drawbtext(menuMiddleRow,1,1);
	drawbtext(menuBottomRow,2,1);
	drawtext(menuFrontRow,2,0);
	if(selectedMain == 1) {
		drawItem(tabTopFile,3,0+32,0);
		drawItem(tabMidFile,3,1+32,0);
		drawItem(tabBotFile,3,2,0);
	} else if (selectedMain == 2) {
		drawItem(tabTopOptions,10,0+32,0);
		drawItem(tabMidOptions,10,1+32,0);
		drawItem(tabBotOptions,10,2,0);
	} else if (selectedMain == 3) {
		drawItem(tabTopAbout,20,0+32,0);
		drawItem(tabMidAbout,20,1+32,0);
		drawItem(tabBotAbout,20,2,0);
	}
}

void uiDummy() {
}

void ui0() {
	changeUI(0);
}
void ui1() {
	changeUI(1);
}
void ui2() {
	changeUI(2);
}
void ui3() {
	changeUI(3);
}
void ui4() {
	changeUI(4);
}
void ui5() {
	changeUI(5);
}
void ui6() {
	changeUI(6);
}
void ui7() {
	changeUI(7);
}
void ui8() {
	changeUI(8);
}
void changeUI(int UInr) {
	selected=0;
	selectedMenu=UInr;
	drawuiX[selectedMenu]();
}


void nullUI() {
	int key;

	key = getinput();
	if( (key&KEY_TOUCH) || ((EMUinput&(KEY_L|KEY_R)) == (KEY_L|KEY_R)) ) {
		selectedMenu = selectedMain = lastMainMenu;
		drawuiX[selectedMenu]();
		powerOn(POWER_2D_B);
		powerOn(PM_BACKLIGHT_BOTTOM);
		if(emuSettings&1) {				// Should we pause when menu is open?
			pauseEmulation = 1;
		}
	}
}

void subUI() {
	int key;

	key=getmenuinput(menuXitems[selectedMenu]);
	if(key&(KEY_A)) {
		if(selectedMain == selectedMenu) {		// are we going from a mainMenu to a subMenu?
			mainUIPosition = selected;
		}
		fnlistX[selectedMenu][selected]();
	}
	if(key&(KEY_B)) {
		lastMainMenu = selectedMain;
		selectedMenu = menuXback[selectedMenu];
		if(selectedMain == selectedMenu) {		// are we going back to a mainMenu?
			selected = mainUIPosition;
		}
	}
	if(key&(KEY_L)) {
		selectedMain--;
		if(selectedMain<=0) {
			lastMainMenu = 1;
			selectedMain = 0;
		}
		selectedMenu = selectedMain;
		selected=0;
	}
	if(key&(KEY_R)) {
		selectedMain++;
		if(selectedMain>3) {
			selectedMain = 3;
		} else {
			selectedMenu = selectedMain;
			selected=0;
		}
	}
	if(key&(KEY_A+KEY_B+KEY_L+KEY_R+KEY_UP+KEY_DOWN+KEY_LEFT+KEY_RIGHT+KEY_TOUCH)) {
		drawuiX[selectedMenu]();
	}
}

void zipTest() {
	char filelist[2048];
	listZipFiles("gberet.zip", filelist);
}

int getmenuinput(int menuitems) {
	int keyhit;

	keyhit = getinput();
	selected = getMenuTouch(&keyhit, selected, menuitems);
	selected = getmenupos(keyhit, selected, menuitems);
	return keyhit;
}

//---------------------------------------------------------------------------------
int getinput() {
	int dpad;
	int keyhit=keysDown();	// buttons pressed this loop

	EMUinput = keysHeld();

	dpad = keysDownRepeat()&(KEY_UP+KEY_DOWN+KEY_LEFT+KEY_RIGHT);
	return dpad|(keyhit&(KEY_A+KEY_B+KEY_START+KEY_L+KEY_R+KEY_TOUCH));
}

//---------------------------------------------------------------------------------
int getmenupos(int keyhit, int sel, int menuitems) {
	if(keyhit&KEY_UP)
		sel=(sel+menuitems-1)%menuitems;
	if(keyhit&KEY_DOWN)
		sel=(sel+1)%menuitems;
	if(keyhit&KEY_RIGHT) {
		sel+=10;
		if(sel>menuitems-1) sel=menuitems-1;
	}
	if(keyhit&KEY_LEFT) {
		sel-=10;
		if(sel<0) sel=0;
	}
	return sel;
}

//---------------------------------------------------------------------------------
int getMenuTouch(int* keyhit, int sel, int menuitems) {
	int ypos, xpos;
	touchPosition myTouch;

	if(*keyhit&KEY_TOUCH) {
		touchRead(&myTouch);
		ypos = (myTouch.py>>3) - 5;
		if( (ypos >= 0) && ((ypos>>1) < menuitems) && !(ypos&1) ) {
			if(sel == (ypos>>1) ) {
				*keyhit |= KEY_A;
			} else {
				sel = ypos>>1;
			}
		} else if(myTouch.py < 24) {
			xpos = myTouch.px>>3;
			if( (xpos > 3) && (xpos < 8)) selectedMenu = selectedMain = 1;
			if( (xpos > 10) && (xpos < 18)) selectedMenu = selectedMain = 2;
			if( (xpos > 20) && (xpos < 26)) selectedMenu = selectedMain = 3;
			if( xpos < 2) *keyhit |= KEY_B;
		}
	}
	return sel;
}

int drawfilelist(const char * dtable, int sel, int items) {
	int i,t,firstitem;
	const char * buf;

	if(sel<12) {
		firstitem=0;
	} else {
		firstitem=sel-11;
		if(sel>(items-13)) {
			firstitem=items-23;
		}
	}

	t = firstitem;
	for(i=0;i<23;i++) {
		buf = dtable + t*64;
		if( *buf == '~' )
			drawtext(buf+1,i+1,i==(sel-firstitem)?1:0);
		else
			drawtext(buf,i+1,i==(sel-firstitem)?1:0);
		t++;
	}
	return (sel-firstitem);
}

void drawtext(const char *str,int row,int hilite) {
	u16 *here=map0sub+row*32;
	int i=0;

	*here=hilite?0x012a:0x0120;
	hilite=(hilite<<12)+0x0100;
//	here++;
	while(str[i]>=' ') {
		here[i]=str[i]|hilite;
		i++;
		if(i>31) break;
	}
	for(;i<32;i++)
		here[i]=0x0120;
}

void drawbtext(const char *str,int row,int shadow) {
	u16 *here=map0sub+(row+32)*32;
	int i=0;

	shadow=(shadow<<13)+0x0100;
	while(str[i]>=' ') {
		here[i]=str[i]|shadow;
		i++;
		if(i>31) break;
	}
	for(;i<32;i++)
		here[i]=0x0120;
}

void strmerge(char *dst,const char *src1,const char *src2) {
	if(dst!=src1)
		strcpy(dst,src1);
	strcat(dst,src2);
}

void drawMenuItem(const char *str1) {
	char str[32];
	if(selected==menuItemRow)
		drawItemBackground(str1, menuItemRow*2+5, 0);
	strmerge(str," ",str1);
	drawtext(str,menuItemRow*2+5,selected==menuItemRow);
	menuItemRow++;
}

void drawSubItem(const char *str1,const char *str2) {
	char str[32];
	if(selected==menuItemRow)
		drawItemBackground(str1, menuItemRow*2+5, 1);
	strmerge(str,"  ",str1);
	strmerge(str,str,str2);
	drawtext(str,menuItemRow*2+5,selected==menuItemRow);
	menuItemRow++;
}

void drawItem(const char *str, int col, int row, int pal) {
	u16 *here=map0sub+col+row*32;
	int attrib = (pal<<12)|0x100;
	int i=0;

	while(str[i]>=' ') {
		here[i] = str[i]|attrib;
		i++;
		if(i>31) break;
	}
}

void drawItemBackground(const char *str, int row, int sub) {
	u16 *here=map0sub+(row+31)*32;
	int i=sub;

	here[i+32] = 0x84|0x2100;
	here[i+64] = 0x86|0x2100;
	here[i++]  = 0x81|0x2100;
	while(str[i]>=' ') {
		here[i+32]= 0x80|0x2100;
		here[i+64]= 0x87|0x2100;
		here[i++] = 0x82|0x2100;
		if(i>31) break;
	}
	here[i+32]= 0x80|0x2100;
	here[i+64]=0x87|0x2100;
	here[i++]=0x82|0x2100;
	here[i+32] = 0x85|0x2100;
	here[i+64] = 0x88|0x2100;
	here[i++] = 0x83|0x2100;
	for(;i<32;i++)
		here[i]=0x0120;
}


void cls(int chrmap) {
	int i=0,len=0x400;				//512x256
	u32 *scr=(u32*)map0sub;
	for(;i<len;i++)
		scr[i]=0x01200120;
}


void resetGame() {
	loadcart(0,0);
}


void autoASet() {
	autoA++;
	joycfg|=KEY_A+(KEY_A<<16);
	if(autoA==1)
		joycfg&=~KEY_A;
	else if(autoA==2)
		joycfg&=~(KEY_A<<16);
	else
		autoA=0;
}

void autoBSet() {
	autoB++;
	joycfg|=KEY_B+(KEY_B<<16);
	if(autoB==1)
		joycfg&=~KEY_B;
	else if(autoB==2)
		joycfg&=~(KEY_B<<16);
	else
		autoB=0;
}

void controllerSet() {					//see io.s: refreshEMUjoypads
	joycfg^=0x20000000;
}

void swapABSet() {
	joycfg^=0x400;
}


void scalingSet(){
	g_scaling^=0x01;
}

void flickSet() {
	g_flicker++;
	if(g_flicker > 1){
		g_flicker=0;
		g_twitch=0;
	}
}

void brightSet() {
	g_gammavalue++;
	if (g_gammavalue>4) g_gammavalue=0;
	paletteInit();
	paletteTxAll();					//make new palette visible
}

void bgrlayerSet(){
	g_gfxmask^=0x03;
}

void sprlayerSet(){
	g_gfxmask^=0x10;
}


void speedSet() {
	int i;
	i = (emuSettings+0x40) & 0xC0;
	emuSettings = (emuSettings & ~0xC0) | i;
	setEmuSpeed(i>>6);
}

void fpsSet() {
	emuSettings = emuSettings^0x08;
}

void autostateSet() {
	emuSettings = emuSettings^0x04;
}

void sleepSet() {
	int i;
	i = (emuSettings+0x10) & 0x30;
	emuSettings = (emuSettings & ~0x30) | i;
	if(i==0x00)
		sleepTime=60*60*5;			// 5min
	else if(i==0x10)
		sleepTime=60*60*10;			// 10min
	else if(i==0x20)
		sleepTime=60*60*30;			// 30min
	else if(i==0x30)
		sleepTime=0x7F000000;		// 360days...
}

void autopauseGameSet() {
	emuSettings^=0x01;
	pauseEmulation = (emuSettings&1);
}

void powerSaveSet() {
	emuSettings^=0x02;
}

void coinaSet() {
	int i;
	i = (g_dipswitch0+1) & 0xF;
	g_dipswitch0 = (g_dipswitch0 & ~0xF) | i;
}
void coinbSet() {
	int i;
	i = (g_dipswitch0+0x10) & 0xF0;
	g_dipswitch0 = (g_dipswitch0 & ~0xF0) | i;
}
void livesSet() {
	int i;
	i = (g_dipswitch1+1) & 3;
	g_dipswitch1 = (g_dipswitch1 & ~3) | i;
}
void bonusSet() {
	int i;
	i = (g_dipswitch1+8) & 0x18;
	g_dipswitch1 = (g_dipswitch1 & ~0x18) | i;
}
void difficultSet() {
	int i;
	i = (g_dipswitch1+0x20) & 0x60;
	g_dipswitch1 = (g_dipswitch1 & ~0x60) | i;
}
void cabinetSet() {
	g_dipswitch1^=0x04;
}
void demoSet() {
	g_dipswitch1^=0x80;
}
void flipSet() {
	g_dipswitch2^=0x01;
}
void uprightSet() {
	g_dipswitch2^=0x02;
}
void serviceSet() {
	g_dipswitch2^=0x04;
}

