#include <nds.h>
#include <fat.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "unzipnds.h"
#include "puff.h"

local_file_hdr ziphead;
central_file_hdr cenhead;
endof_central_hdr endhead;
char zip_err[40];

// Returns offset of file in zipfile. Error if val < 0.
int locateFileInZip(const char* filename, FILE* myfile) {
	int retval=0;
	char txtbuf[256];

	fseek(myfile, - ENDHEADLEN, SEEK_END);
	endhead.magic = 0;
	fread(&endhead, 1, ENDHEADLEN, myfile);
	if(	endhead.magic == ENDSIG) {
		fseek(myfile, endhead.cdir_offset, SEEK_SET);

		while(1){
			fillCentralZipHeader(&cenhead, myfile);
			if(	cenhead.magic == CENSIG) {
				fread(txtbuf, 1, cenhead.filename_length, myfile);
				txtbuf[cenhead.filename_length] = 0;
				if(strstr(txtbuf, filename)) {
					retval = cenhead.relative_offset_local_header;
					break;
				}
				fseek(myfile, cenhead.extra_field_length + cenhead.file_comment_length, SEEK_CUR);
			} else {
				retval = -1;		// File not found
				break;
			}
		}
	} else {
		retval = -1;		// File not found
	}

	return retval;
}

int findFileInZip(const char* zipname, const char* filename) {
	int err=0;
	FILE* myfile;

	if( (myfile = fopen(zipname, "r")) ) {
		if( (locateFileInZip(filename, myfile)) < 0)
			err = 1;		// File not found in zip
		fclose(myfile);
	} else {
		err = 1;		// ZipFile not found
	}
	return err;
}

int loadFileInZip(void* dest, const char* zipname, const char* filename) {
	int err=0, offset=0, csize, ucsize;
	FILE* myfile;
	unsigned char* src;

	if( (myfile = fopen(zipname, "r")) ) {
		offset = locateFileInZip(filename, myfile);
		if( offset >= 0) {
			fseek(myfile, offset, SEEK_SET);
			fillLocalZipHeader(&ziphead, myfile);
			csize = cenhead.csize;							// localhead doesn't allways contain all info.
			ucsize = cenhead.ucsize;						// centralhead should allways be available.
			if(	ziphead.magic == LOCSIG) {
				if( ziphead.compression_method == 8 ) {		// 8 == DEFLATE compression.
					if( (src = malloc(csize)) ) {
						fseek(myfile, ziphead.extra_field_length + ziphead.filename_length, SEEK_CUR);
						fread(src, 1, csize, myfile);
						puff(dest, (unsigned long*)&ucsize, src, (unsigned long*)&ucsize);
						free(src);
					} else {
						strncpy(zip_err,"Out of memory!", 32);
						err = 1;
					}
				} else if( ziphead.compression_method == 0 ) {		// 0 == STORE, no compression.
					fseek(myfile, ziphead.extra_field_length, SEEK_CUR);
					fread(dest, 1, ziphead.ucsize, myfile);
				} else {									// Only supports DEFLATE & STORE.
					strncpy(zip_err,"Unsupported compression in zip.", 32);
					err = 1;
				}
			} else {
				strncpy(zip_err,"Couldn't find file in zip.", 32);
				err = 1;
			}
		} else {
			strncpy(zip_err,"Couldn't find file in zip.", 32);
			err = 1;
		}
		fclose(myfile);
	} else {
		strncpy(zip_err,"Couldn't open zip file.", 32);
		err = 1;
	}

	return err;
}

int listZipFiles(const char* zipname, char* table) {
	int len;
	FILE* myfile;
	char txtbuf[256];

	if( (myfile = fopen(zipname, "r")) ) {
		while(1){
			fillLocalZipHeader(&ziphead, myfile);
			if(	ziphead.magic == LOCSIG) {
				fread(txtbuf, 1, ziphead.filename_length, myfile);
				txtbuf[ziphead.filename_length] = 0;
				strcat(table, txtbuf);
				len = ziphead.extra_field_length + ziphead.csize;
				fseek(myfile, len, SEEK_CUR);
			} else {
				break;
			}
		}
		fclose(myfile);
	} else {
		return 1;
	}
	return 0;
}

int fillLocalZipHeader(local_file_hdr* hdr, FILE* myfile) {
	hdr->magic = 0;
	fread(&hdr->magic, 1, sizeof(u32), myfile);
	fread(&hdr->version_needed_to_extract, 1, sizeof(u16), myfile);
	fread(&hdr->general_purpose_bit_flag, 1, sizeof(u16), myfile);
	fread(&hdr->compression_method, 1, sizeof(u16), myfile);
	fread(&hdr->last_mod_dos_datetime, 1, sizeof(u32), myfile);
	fread(&hdr->crc32, 1, sizeof(u32), myfile);
	fread(&hdr->csize, 1, sizeof(u32), myfile);
	fread(&hdr->ucsize, 1, sizeof(u32), myfile);
	fread(&hdr->filename_length, 1, sizeof(u16), myfile);
	fread(&hdr->extra_field_length, 1, sizeof(u16), myfile);
	return 0;
}

int fillCentralZipHeader(central_file_hdr* hdr, FILE* myfile) {
	hdr->magic = 0;
	fread(&hdr->magic, 1, sizeof(u32), myfile);
	fread(&hdr->version_made_by, 1, sizeof(u16), myfile);
	fread(&hdr->version_needed_to_extract, 1, sizeof(u16), myfile);
	fread(&hdr->general_purpose_bit_flag, 1, sizeof(u16), myfile);
	fread(&hdr->compression_method, 1, sizeof(u16), myfile);
	fread(&hdr->last_mod_dos_datetime, 1, sizeof(u32), myfile);
	fread(&hdr->crc32, 1, sizeof(u32), myfile);
	fread(&hdr->csize, 1, sizeof(u32), myfile);
	fread(&hdr->ucsize, 1, sizeof(u32), myfile);
	fread(&hdr->filename_length, 1, sizeof(u16), myfile);
	fread(&hdr->extra_field_length, 1, sizeof(u16), myfile);
	fread(&hdr->file_comment_length, 1, sizeof(u16), myfile);
	fread(&hdr->disk_number_start, 1, sizeof(u16), myfile);
	fread(&hdr->internal_file_attributes, 1, sizeof(u16), myfile);
	fread(&hdr->external_file_attributes, 1, sizeof(u32), myfile);
	fread(&hdr->relative_offset_local_header, 1, sizeof(u32), myfile);
	return 0;
}
